/* 

   March 2009, Ferragina Paolo (ferragina@di.unipi.it)
   bmiu.c -- My adaptation of the Bentley-McIlroy's decompressor

   Code originally downloaded from: www.cs.dartmouth.edu/~doug/source.html

   My adaptation consists of porting to the GZIP-like interface, remove
   some specialties that made the code a little bit cumbersome (for
   me, at least!), change the copy-encoding to a pair of 4-bytes ints,
   and changing the hash-table implementation to make it more
   performing (TBD...). 

   NOTICE: This code can be used only for research purposes. All rights
   are owned by the author: Paolo Ferragina, Dipartimento di Informatica,
   University of Pisa, Italy.

*/


#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

typedef unsigned long uLong;


// Simple fixed-len encoding on 4 bytes
int getVarInt(int numBytes, unsigned char *a, int *n) 
{	
  int c, i;

  c=0;
  for(i=0; i < numBytes; i++){
    c <<= 8;
    c += (((int) a[*n]) & 0xFF);
    *n += 1;
  }

  return c;
}


void bmi_decompress(unsigned char *outBox, int *outBoxLen, unsigned char *inBox, int inBoxLen)
{	
  int i,in,outpos, c, start, len;

  outpos=0;

  for(in=0; in < inBoxLen; ){
    c = inBox[in++];

    if (c == '<') {
      c = inBox[in++]; /* danger: no EOF */

      if (c == 0) { /* quoted < */
	outBox[outpos++] = '<';
      } else { // copy instr

	if(c > 16){
	  printf("Error in decoding dist-pair len (%d)! \n",c);
	  exit(1); }

	int bytesDist = c / 4;
	int bytesLen = c % 4;
	  
	start = getVarInt(bytesDist,inBox,&in);
	start = outpos - start; 
	len = getVarInt(bytesLen,inBox,&in);

	for (i = start; i < start+len; i++){
	  outBox[outpos++] = outBox[i];
	}
      }

    } else { // no symbol <, just copy
      outBox[outpos++] = c;
    }
  }

  if(outpos >= *outBoxLen){
    printf("Error in BMI decompress (%d/%d)\n\n", outpos,*outBoxLen);
    exit(1);
  }

  *outBoxLen = outpos;
}


